//! Bundle all functions together.

use std::{
    io,
    path::{Path, PathBuf},
};

use r2d2::Pool;
use r2d2_sqlite::SqliteConnectionManager;
use tempfile::{NamedTempFile, PersistError};

use crate::{
    data::{self, author::Author, book::Book, pagination::SortOrder, series::Series},
    search::search,
};
use snafu::{ResultExt, Snafu};

/// Top level calibre functions, bundling all sub functions in one place and providing secure access to
/// the database.
#[derive(Debug, Clone)]
pub struct Calibre {
    pool: Pool<SqliteConnectionManager>,
    search_db_path: PathBuf,
}

#[derive(Debug, Snafu)]
pub enum LoadError {
    #[snafu(display("Failed to create database connection pool."))]
    DbPool { source: r2d2::Error },
    #[snafu(display("Failed to create temporary database view."))]
    TmpDb { source: io::Error },
    #[snafu(display("Failed to keep temporary database from deletion."))]
    PersistTmpDb { source: PersistError },
}

#[derive(Debug, Snafu)]
#[snafu(display("Failed to search."))]
pub struct SearchError {
    source: crate::search::SearchError,
}

#[derive(Debug, Snafu)]
pub enum BooksError {
    #[snafu(display("Failed to get database connection from pool."))]
    BooksDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch multiple books."))]
    FetchBooks {
        source: data::book::MultipleBooksError,
    },
}

#[derive(Debug, Snafu)]
pub enum AuthorsError {
    #[snafu(display("Failed to get database connection from pool."))]
    AuthorsDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch multiple authors."))]
    FetchAuthors {
        source: data::author::MultipleAuthorsError,
    },
}

#[derive(Debug, Snafu)]
pub enum AuthorBooksError {
    #[snafu(display("Failed to get database connection from pool."))]
    AuthorBooksDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch multiple  an author's books."))]
    FetchAuthorBooks {
        source: data::book::AuthorBooksError,
    },
}

#[derive(Debug, Snafu)]
pub enum RecentBooksError {
    #[snafu(display("Failed to get database connection from pool."))]
    RecentBooksDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch recent books."))]
    FetchRecentBooks {
        source: data::book::RecentBooksError,
    },
}

#[derive(Debug, Snafu)]
pub enum ScalarBookError {
    #[snafu(display("Failed to get database connection from pool."))]
    ScalarBookDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch book."))]
    FetchScalarBook { source: data::book::ScalarBookError },
}

#[derive(Debug, Snafu)]
pub enum ScalarAuthorError {
    #[snafu(display("Failed to get database connection from pool."))]
    ScalarAuthorDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch author."))]
    FetchScalarAuthor {
        source: data::author::ScalarAuthorError,
    },
}

#[derive(Debug, Snafu)]
pub enum ScalarSeriesError {
    #[snafu(display("Failed to get database connection from pool."))]
    ScalarSeriesDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch series."))]
    FetchScalarSeries {
        source: data::series::ScalarSeriesError,
    },
}

#[derive(Debug, Snafu)]
pub enum BookAuthorError {
    #[snafu(display("Failed to get database connection from pool."))]
    BookAuthorDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch book's author."))]
    FetchBookAuthor {
        source: data::author::BookAuthorError,
    },
}

#[derive(Debug, Snafu)]
pub enum MultipleSeriesError {
    #[snafu(display("Failed to get database connection from pool."))]
    MultipleSeriesDbPool { source: r2d2::Error },
    #[snafu(display("Failed to fetch multiple series."))]
    FetchMultipleSeries {
        source: data::series::MultiplSeriesError,
    },
}

#[derive(Debug, Snafu)]
pub enum BookSeriesError {
    #[snafu(display("Failed to get database connection from pool."))]
    BookSeriesDbPool { source: r2d2::Error },
    #[snafu(display("Failed to get the series a book belongs to."))]
    FetchBookSeries {
        source: data::series::SeriesBooksError,
    },
}

#[derive(Debug, Snafu)]
pub enum SeriesBooksError {
    #[snafu(display("Failed to get database connection from pool."))]
    SeriesBooksDbPool { source: r2d2::Error },
    #[snafu(display("Failed to get a series' books."))]
    FetchSeriesBooks { source: data::book::SeriesBookError },
}

#[derive(Debug, Snafu)]
pub enum HasPreviousAuthorsError {
    #[snafu(display("Failed to get database connection from pool."))]
    HasPreviousAuthorsDbPool { source: r2d2::Error },
    #[snafu(display("Failed to check if there are previous authors."))]
    FetchHasPreviousAuthors {
        source: data::author::PreviousAuthorsError,
    },
}

#[derive(Debug, Snafu)]
pub enum HasMoreAuthorsError {
    #[snafu(display("Failed to get database connection from pool."))]
    HasMoreAuthorsDbPool { source: r2d2::Error },
    #[snafu(display("Failed to check if there are previous authors."))]
    FetchHasMoreAuthors {
        source: data::author::MoreAuthorsError,
    },
}

#[derive(Debug, Snafu)]
pub enum HasPreviousBooksError {
    #[snafu(display("Failed to get database connection from pool."))]
    HasPreviousBooksDbPool { source: r2d2::Error },
    #[snafu(display("Failed to check if there are previous books."))]
    FetchHasPreviousBooks {
        source: data::book::PreviousBooksError,
    },
}

#[derive(Debug, Snafu)]
pub enum HasMoreBooksError {
    #[snafu(display("Failed to get database connection from pool."))]
    HasMoreBooksDbPool { source: r2d2::Error },
    #[snafu(display("Failed to check if there are previous books."))]
    FetchHasMoreBooks { source: data::book::MoreBooksError },
}

#[derive(Debug, Snafu)]
pub enum HasPreviousSeriesError {
    #[snafu(display("Failed to get database connection from pool."))]
    HasPreviousSeriesDbPool { source: r2d2::Error },
    #[snafu(display("Failed to check if there are previous series."))]
    FetchHasPreviousSeries {
        source: data::series::PreviousSeriesError,
    },
}

#[derive(Debug, Snafu)]
pub enum HasMoreSeriesError {
    #[snafu(display("Failed to get database connection from pool."))]
    HasMoreSeriesDbPool { source: r2d2::Error },
    #[snafu(display("Failed to check if there are previous series."))]
    FetchHasMoreSeries {
        source: data::series::MoreSeriesError,
    },
}
impl Calibre {
    /// Open a connection to the calibre database.
    ///
    /// Fail if the database file can not be opened or not be found.
    pub fn load(path: &Path) -> Result<Self, LoadError> {
        let manager = SqliteConnectionManager::file(path);
        let pool = r2d2::Pool::new(manager).context(DbPoolSnafu)?;

        let tmpfile = NamedTempFile::new().context(TmpDbSnafu)?;
        let (_, search_db_path) = tmpfile.keep().context(PersistTmpDbSnafu)?;

        Ok(Self {
            pool,
            search_db_path,
        })
    }

    /// Full text search with a query.
    ///
    /// See https://www.sqlite.org/fts5.html#full_text_query_syntax for syntax.
    pub fn search(&self, query: &str) -> Result<Vec<Book>, SearchError> {
        search(query, &self.pool, &self.search_db_path).context(SearchSnafu)
    }

    /// Fetch book data from calibre, starting at `cursor`, fetching up to an amount of `limit` and
    /// ordering by `sort_order`.
    pub fn books(
        &self,
        limit: u64,
        cursor: Option<&str>,
        sort_order: &SortOrder,
    ) -> Result<Vec<Book>, BooksError> {
        let conn = self.pool.get().context(BooksDbPoolSnafu)?;
        Book::multiple(&conn, limit, cursor, sort_order).context(FetchBooksSnafu)
    }

    /// Fetch author data from calibre, starting at `cursor`, fetching up to an amount of `limit` and
    /// ordering by `sort_order`.
    pub fn authors(
        &self,
        limit: u64,
        cursor: Option<&str>,
        sort_order: &SortOrder,
    ) -> Result<Vec<Author>, AuthorsError> {
        let conn = self.pool.get().context(AuthorsDbPoolSnafu)?;
        Author::multiple(&conn, limit, cursor, sort_order).context(FetchAuthorsSnafu)
    }

    /// Fetch books for an author specified by `author_id`, paginate the books by starting at `cursor`,
    /// fetching up to an amount of `limit` and ordering by `sort_order`.
    pub fn author_books(
        &self,
        author_id: u64,
        limit: u64,
        cursor: Option<&str>,
        sort_order: SortOrder,
    ) -> Result<Vec<Book>, AuthorBooksError> {
        let conn = self.pool.get().context(AuthorBooksDbPoolSnafu)?;
        Book::author_books(&conn, author_id, limit, cursor, sort_order)
            .context(FetchAuthorBooksSnafu)
    }

    /// Get recent books up to a limit of `limit`.
    pub fn recent_books(&self, limit: u64) -> Result<Vec<Book>, RecentBooksError> {
        let conn = self.pool.get().context(RecentBooksDbPoolSnafu)?;
        Book::recents(&conn, limit).context(FetchRecentBooksSnafu)
    }

    /// Get a single book, specified `id`.
    pub fn scalar_book(&self, id: u64) -> Result<Book, ScalarBookError> {
        let conn = self.pool.get().context(ScalarBookDbPoolSnafu)?;
        Book::scalar_book(&conn, id).context(FetchScalarBookSnafu)
    }

    /// Get the author to a book with id `id`.
    pub fn book_author(&self, id: u64) -> Result<Author, BookAuthorError> {
        let conn = self.pool.get().context(BookAuthorDbPoolSnafu)?;
        Author::book_author(&conn, id).context(FetchBookAuthorSnafu)
    }

    /// Fetch series data from calibre, starting at `cursor`, fetching up to an amount of `limit` and
    /// ordering by `sort_order`.
    pub fn series(
        &self,
        limit: u64,
        cursor: Option<&str>,
        sort_order: &SortOrder,
    ) -> Result<Vec<Series>, MultipleSeriesError> {
        let conn = self.pool.get().context(MultipleSeriesDbPoolSnafu)?;
        Series::multiple(&conn, limit, cursor, sort_order).context(FetchMultipleSeriesSnafu)
    }

    /// Get the series a book with id `id` is in, as well as the book's position within the series.
    pub fn book_series(&self, id: u64) -> Result<Option<(Series, f64)>, BookSeriesError> {
        let conn = self.pool.get().context(BookSeriesDbPoolSnafu)?;
        Series::book_series(&conn, id).context(FetchBookSeriesSnafu)
    }

    /// Get all books belonging to the series with id `id`.
    pub fn series_books(&self, id: u64) -> Result<Vec<Book>, SeriesBooksError> {
        let conn = self.pool.get().context(SeriesBooksDbPoolSnafu)?;
        Book::series_books(&conn, id).context(FetchSeriesBooksSnafu)
    }

    /// Check if there are more authors before the specified cursor.
    pub fn has_previous_authors(&self, author_sort: &str) -> Result<bool, HasPreviousAuthorsError> {
        let conn = self.pool.get().context(HasPreviousAuthorsDbPoolSnafu)?;
        Author::has_previous_authors(&conn, author_sort).context(FetchHasPreviousAuthorsSnafu)
    }

    /// Check if there are more authors after the specified cursor.
    pub fn has_more_authors(&self, author_sort: &str) -> Result<bool, HasMoreAuthorsError> {
        let conn = self.pool.get().context(HasMoreAuthorsDbPoolSnafu)?;
        Author::has_more_authors(&conn, author_sort).context(FetchHasMoreAuthorsSnafu)
    }

    /// Check if there are more books before the specified cursor.
    pub fn has_previous_books(&self, book_sort: &str) -> Result<bool, HasPreviousBooksError> {
        let conn = self.pool.get().context(HasPreviousBooksDbPoolSnafu)?;
        Book::has_previous_books(&conn, book_sort).context(FetchHasPreviousBooksSnafu)
    }

    /// Check if there are more books after the specified cursor.
    pub fn has_more_books(&self, book_sort: &str) -> Result<bool, HasMoreBooksError> {
        let conn = self.pool.get().context(HasMoreBooksDbPoolSnafu)?;
        Book::has_more_books(&conn, book_sort).context(FetchHasMoreBooksSnafu)
    }

    /// Check if there are more series before the specified cursor.
    pub fn has_previous_series(&self, series_sort: &str) -> Result<bool, HasPreviousSeriesError> {
        let conn = self.pool.get().context(HasPreviousSeriesDbPoolSnafu)?;
        Series::has_previous_series(&conn, series_sort).context(FetchHasPreviousSeriesSnafu)
    }

    /// Check if there are more series after the specified cursor.
    pub fn has_more_series(&self, series_sort: &str) -> Result<bool, HasMoreSeriesError> {
        let conn = self.pool.get().context(HasMoreSeriesDbPoolSnafu)?;
        Series::has_more_series(&conn, series_sort).context(FetchHasMoreSeriesSnafu)
    }

    /// Fetch a single author with id `id`.
    pub fn scalar_author(&self, id: u64) -> Result<Author, ScalarAuthorError> {
        let conn = self.pool.get().context(ScalarAuthorDbPoolSnafu)?;
        Author::scalar_author(&conn, id).context(FetchScalarAuthorSnafu)
    }

    /// Fetch a single series with id `id`.
    pub fn scalar_series(&self, id: u64) -> Result<Series, ScalarSeriesError> {
        let conn = self.pool.get().context(ScalarSeriesDbPoolSnafu)?;
        Series::scalar_series(&conn, id).context(FetchScalarSeriesSnafu)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    fn init_calibre() -> Calibre {
        Calibre::load(Path::new("./testdata/metadata.db")).unwrap()
    }

    #[test]
    fn books() {
        let c = init_calibre();
        let books = c.books(10, None, &SortOrder::ASC).unwrap();
        assert_eq!(books.len(), 4);
    }

    #[test]
    fn authors() {
        let c = init_calibre();
        let authors = c.authors(10, None, &SortOrder::ASC).unwrap();
        assert_eq!(authors.len(), 3);
    }

    #[test]
    fn author_books() {
        let c = init_calibre();
        let books = c.author_books(1, 10, None, SortOrder::ASC).unwrap();
        assert_eq!(books.len(), 2);
    }

    #[test]
    fn pagination() {
        let c = init_calibre();
        let authors = c.authors(1, None, &SortOrder::ASC).unwrap();
        assert_eq!(authors.len(), 1);
        assert_eq!(authors[0].name, "Kevin R. Grazier");

        let authors = c
            .authors(1, Some(&authors[0].sort), &SortOrder::ASC)
            .unwrap();
        assert_eq!(authors.len(), 1);
        assert_eq!(authors[0].name, "Terry Pratchett");

        let authors = c
            .authors(1, Some(&authors[0].sort), &SortOrder::ASC)
            .unwrap();
        assert_eq!(authors.len(), 1);
        assert_eq!(authors[0].name, "Edward Noyes Westcott");

        let authors = c
            .authors(1, Some(&authors[0].sort), &SortOrder::DESC)
            .unwrap();
        assert_eq!(authors.len(), 1);
        assert_eq!(authors[0].name, "Terry Pratchett");
    }
}
