//! Cursor pagination handling.

use rusqlite::{Connection, Row, ToSql, named_params};
use serde::{Deserialize, Serialize};
use snafu::{ResultExt, Snafu};

/// How to sort query results. Signifying whether we are paginating forwards or backwards.
#[derive(Debug, Copy, Clone, PartialEq, Deserialize, Serialize)]
pub enum SortOrder {
    /// Forwards
    ASC,
    /// Backwards
    DESC,
}

/// Pagination data.
pub struct Pagination<'a> {
    /// Sort by this column.
    pub sort_col: &'a str,
    /// Limit returned results.
    pub limit: u64,
    /// Where to start paginating.
    pub cursor: Option<&'a str>,
    /// Paginating forwards or backwards.
    pub sort_order: SortOrder,
}

#[derive(Debug, Snafu)]
pub enum HasPrevOrMoreError {
    #[snafu(display("Failed to prepare statement."))]
    PrepareHasPrevOrMore { source: rusqlite::Error },
    #[snafu(display("Failed to execute statement."))]
    ExecuteHasPrevOrMore { source: rusqlite::Error },
}

#[derive(Debug, Snafu)]
pub enum PaginationError {
    #[snafu(display("Failed to prepare statement."))]
    PreparePagination { source: rusqlite::Error },
    #[snafu(display("Failed to execute statement."))]
    ExecutePagination { source: rusqlite::Error },
}

impl<'a> Pagination<'a> {
    /// Create a new pagination.
    pub fn new(
        sort_col: &'a str,
        cursor: Option<&'a str>,
        limit: u64,
        sort_order: SortOrder,
    ) -> Self {
        Self {
            sort_col,
            limit,
            cursor,
            sort_order,
        }
    }

    fn sort_order_to_sql(sort_order: &SortOrder) -> String {
        if *sort_order == SortOrder::ASC {
            ">"
        } else {
            "<"
        }
        .to_string()
    }

    /// Check if there are more items forwards or backwards from `cursor` (direction specified by `sort_order`).
    pub fn has_prev_or_more(
        conn: &Connection,
        table: &str,
        sort: &str,
        sort_order: &SortOrder,
    ) -> Result<bool, HasPrevOrMoreError> {
        let comparison = Pagination::sort_order_to_sql(sort_order);

        let mut stmt = conn.prepare(&format!(
            "SELECT Count(1) FROM {table} WHERE sort {comparison} (:sort) ORDER BY sort {sort_order:?}"
        )).context(PrepareHasPrevOrMoreSnafu)?;
        let params = named_params! { ":sort": sort};
        let count: u64 = stmt
            .query_row(params, |x| x.get(0))
            .context(ExecuteHasPrevOrMoreSnafu)?;

        Ok(count > 0)
    }

    /// Paginate a statement.
    pub fn paginate<T, F>(
        &self,
        conn: &Connection,
        statement: &str,
        params: &[(&str, &dyn ToSql)],
        processor: F,
    ) -> Result<Vec<T>, PaginationError>
    where
        F: FnMut(&Row<'_>) -> Result<T, rusqlite::Error>,
    {
        let cursor = self.cursor.unwrap_or("");
        let comparison = Pagination::sort_order_to_sql(&self.sort_order);

        let where_sql = if statement.ends_with("AND") {
            ""
        } else {
            "WHERE"
        };

        let sort_col = self.sort_col;
        // otherwise paginated statements with join will fails, not happy with this but fine for
        // now
        let sort_col_wrapped = if let Some(index) = sort_col.find('.') {
            let right_part = &sort_col[index..];
            "t".to_owned() + right_part
        } else {
            sort_col.to_owned()
        };
        let sort_order = &self.sort_order;
        // DANGER: vulnerable to SQL injection if statement or sort_col variable is influenced by user input
        let mut stmt = conn.prepare(&format!(
            "SELECT * FROM ({statement} {where_sql} {sort_col} {comparison} (:cursor) ORDER BY {sort_col} {sort_order:?} LIMIT (:limit)) AS t ORDER BY {sort_col_wrapped} ASC"
        )).context(PreparePaginationSnafu)?;
        let params = [
            &[
                (":cursor", &cursor as &dyn ToSql),
                (":limit", &self.limit as &dyn ToSql),
            ],
            params,
        ]
        .concat();
        let iter = stmt
            .query_map(params.as_slice(), processor)
            .context(ExecutePaginationSnafu)?;
        Ok(iter.filter_map(Result::ok).collect())
    }
}
