use std::{io, net::SocketAddr};

use serde::Deserialize;
use snafu::{ResultExt, Snafu};
use tokio::net::TcpListener;
use utoipa::{OpenApi, ToSchema};
use utoipa_axum::router::OpenApiRouter;
use utoipa_swagger_ui::SwaggerUi;

use crate::app_state::AppState;

pub mod authors;
pub mod books;
pub mod download;
pub mod error;
pub mod html;
pub mod opds;
pub mod paginated;
pub mod routes;
pub mod search;
pub mod series;
pub mod static_files;

/// How to sort query results.
#[derive(Debug, Copy, Clone, PartialEq, Deserialize, ToSchema)]
#[serde(rename_all = "UPPERCASE")]
pub enum SortOrder {
    ASC,
    DESC,
}

impl From<SortOrder> for calibre_db::data::pagination::SortOrder {
    fn from(val: SortOrder) -> Self {
        match val {
            SortOrder::ASC => calibre_db::data::pagination::SortOrder::ASC,
            SortOrder::DESC => calibre_db::data::pagination::SortOrder::DESC,
        }
    }
}

/// OpenAPI tag for all endpoints.
const TAG: &str = "little-hesinde";

/// OpenAPI documentation configuration.
#[derive(OpenApi)]
#[openapi(
      components(
            schemas(
                SortOrder
            )
        ),
        tags(
            (name = TAG, description = "Browser and OPDS access to a calibre library.")
        )
    )]
struct ApiDoc;

/// Errors that occur when starting the HTTP server.
#[derive(Debug, Snafu)]
pub enum ServeError {
    #[snafu(display("Failed to bind to {address}."))]
    Bind {
        source: io::Error,
        address: SocketAddr,
    },
    #[snafu(display("Failed to run http server."))]
    Serve { source: io::Error },
}

/// Start the HTTP API server with the given configuration.
pub async fn serve(address: SocketAddr, state: AppState) -> Result<(), ServeError> {
    let (router, api) = OpenApiRouter::with_openapi(ApiDoc::openapi())
        .merge(routes::router(state))
        .split_for_parts();

    let router =
        router.merge(SwaggerUi::new("/swagger-ui").url("/api-docs/openapi.json", api.clone()));

    let listener = TcpListener::bind(&address)
        .await
        .context(BindSnafu { address })?;

    axum::serve(listener, router.into_make_service())
        .await
        .context(ServeSnafu)
}
