//! HTTP error handling and response formatting.

use axum::http::StatusCode;
use serde::Serialize;
use utoipa::ToSchema;

/// Standard error response format for API endpoints.
#[derive(Serialize, ToSchema)]
pub struct ErrorResponse {
    /// Unique identifier for tracking this error instance.
    pub id: String,
    /// Human-readable error message.
    pub error: String,
}

/// Map error types to HTTP status codes.
pub trait HttpStatus {
    /// Return the appropriate HTTP status code for this error.
    fn status_code(&self) -> StatusCode;
}

/// Generate IntoResponse implementation for error types with JSON formatting.
#[macro_export]
macro_rules! http_error {
    ($error_type:ty) => {
        impl axum::response::IntoResponse for $error_type {
            fn into_response(self) -> axum::response::Response {
                let status = self.status_code();
                let id = uuid::Uuid::new_v4().to_string();
                tracing::error!("{}: {}", &id, snafu::Report::from_error(&self));

                let error_response = $crate::api::error::ErrorResponse {
                    id,
                    error: self.to_string(),
                };

                (status, axum::Json(error_response)).into_response()
            }
        }
    };
}
