use axum::{http::StatusCode, response::Response};
use snafu::{ResultExt, Snafu};

use crate::{
    APP_NAME, APP_VERSION,
    api::{
        TAG,
        download::{self, DownloadError},
        error::{ErrorResponse, HttpStatus},
    },
    http_error,
};

const SOURCE_ARCHIVE: &[u8] = include_bytes!(concat!(env!("OUT_DIR"), "/archive.zip"));

#[derive(Debug, Snafu)]
pub enum ArchiveError {
    #[snafu(display("Failed to stream source code archive."))]
    Download { source: DownloadError },
}
impl HttpStatus for ArchiveError {
    fn status_code(&self) -> StatusCode {
        match self {
            ArchiveError::Download { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(ArchiveError);

/// Handle a request for source code of the server..
#[utoipa::path(
    get,
    path = "/archive",
    tag = TAG,
    responses(
        (status = OK, content_type = "application/zip"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler() -> Result<Response, ArchiveError> {
    let file_name = format!("{APP_NAME}-{APP_VERSION}.zip");
    download::handler(&file_name, SOURCE_ARCHIVE, "application/zip")
        .await
        .context(DownloadSnafu)
}
