use std::sync::Arc;

use crate::{
    api::{
        SortOrder, TAG,
        authors::{self, SingleAuthorError},
        error::{ErrorResponse, HttpStatus},
        paginated::{self, PaginationError},
    },
    app_state::AppState,
    http_error,
    templates::TEMPLATES,
};
use axum::{
    extract::{Path, State},
    http::StatusCode,
    response::{Html, IntoResponse, Response},
};
use snafu::{ResultExt, Snafu};
use tera::Context;

#[derive(Debug, Snafu)]
pub enum RetrieveError {
    #[snafu(display("Failed to fetch pagination data."))]
    Authors { source: AuthorError },
}
impl HttpStatus for RetrieveError {
    fn status_code(&self) -> StatusCode {
        match self {
            RetrieveError::Authors { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(RetrieveError);

#[utoipa::path(
    get,
    path = "/authors",
    tag = TAG,
    responses(
        (status = OK, content_type = "text/html"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler_init(State(state): State<Arc<AppState>>) -> Result<Response, RetrieveError> {
    authors(&state, None, SortOrder::ASC)
        .await
        .context(AuthorsSnafu)
}

#[utoipa::path(
    get,
    path = "/authors/{cursor}/{sort_order}",
    tag = TAG,
    responses(
        (status = OK, content_type = "text/html"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler(
    Path((cursor, sort_order)): Path<(String, SortOrder)>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, RetrieveError> {
    authors(&state, Some(&cursor), sort_order)
        .await
        .context(AuthorsSnafu)
}

#[derive(Debug, Snafu)]
pub enum AuthorError {
    #[snafu(display("Failed to fetch pagination data."))]
    Pagination { source: PaginationError },
}

async fn authors(
    state: &Arc<AppState>,
    cursor: Option<&str>,
    sort_order: SortOrder,
) -> Result<Response, AuthorError> {
    paginated::render(
        "authors",
        || state.calibre.authors(25, cursor, &sort_order.into()),
        |author| author.sort.clone(),
        |cursor| state.calibre.has_previous_authors(cursor),
        |cursor| state.calibre.has_more_authors(cursor),
    )
    .context(PaginationSnafu)
}

#[derive(Debug, Snafu)]
pub enum SingleError {
    #[snafu(display("Failed to fetch author data."))]
    Data { source: SingleAuthorError },
    #[snafu(display("Failed to render template."))]
    Render { source: tera::Error },
}
impl HttpStatus for SingleError {
    fn status_code(&self) -> StatusCode {
        match self {
            SingleError::Data { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            SingleError::Render { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(SingleError);

#[utoipa::path(
    get,
    path = "/authors/{id}",
    tag = TAG,
    responses(
        (status = OK, content_type = "text/html"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn single(
    Path(id): Path<u64>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, SingleError> {
    let (author, books) = authors::single(id, &state.calibre, &state.config.library_path)
        .await
        .context(DataSnafu)?;

    let mut context = Context::new();
    context.insert("title", &author.name);
    context.insert("nav", "authors");
    context.insert("books", &books);

    Ok(TEMPLATES
        .render("book_list", &context)
        .context(RenderSnafu)
        .map(Html)?
        .into_response())
}
