use std::sync::Arc;

use axum::{
    extract::{Query, State},
    http::StatusCode,
    response::{Html, IntoResponse, Response},
};
use serde::Deserialize;
use snafu::{ResultExt, Snafu};
use tera::Context;

use crate::{
    api::{
        TAG,
        error::{ErrorResponse, HttpStatus},
        search::{self, SearchQueryError},
    },
    app_state::AppState,
    http_error,
    templates::TEMPLATES,
};

/// Errors that occur during query processing.
#[derive(Debug, Snafu)]
pub enum SearchError {
    #[snafu(display("Failed to search for books."))]
    Query { source: SearchQueryError },
    #[snafu(display("Failed to render template."))]
    Template { source: tera::Error },
}
impl HttpStatus for SearchError {
    fn status_code(&self) -> StatusCode {
        match self {
            SearchError::Query { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            SearchError::Template { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(SearchError);

#[derive(Deserialize)]
pub struct Params {
    /// Query for a search request.
    query: String,
}
#[utoipa::path(
    get,
    path = "/search",
    tag = TAG,
    responses(
        (status = 200, content_type = "text/html"),
        (status = 500, description = "Error retrieving books from database", body = ErrorResponse)
    )
)]
pub async fn handler(
    Query(params): Query<Params>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, SearchError> {
    let books = search::query(&params.query, &state.calibre, &state.config.library_path)
        .await
        .context(QuerySnafu)?;

    let mut context = Context::new();
    context.insert("title", "Search Results");
    context.insert("nav", "search");
    context.insert("books", &books);

    Ok(TEMPLATES
        .render("book_list", &context)
        .context(TemplateSnafu)
        .map(Html)?
        .into_response())
}
