use std::sync::Arc;

use crate::{
    APP_NAME,
    api::{
        SortOrder, TAG,
        authors::{self, SingleAuthorError},
        error::{ErrorResponse, HttpStatus},
    },
    app_state::AppState,
    http_error,
    opds::{
        entry::Entry, error::AsXmlError, feed::Feed, link::Link, media_type::MediaType,
        relation::Relation,
    },
};
use axum::{
    extract::{Path, State},
    http::{StatusCode, header},
    response::{IntoResponse, Response},
};
use calibre_db::{calibre::AuthorsError, data::author::Author as DbAuthor};
use snafu::{ResultExt, Snafu};
use time::OffsetDateTime;

#[derive(Debug, Snafu)]
pub enum AuthorsError {
    #[snafu(display("Failed to fetch author data."))]
    Data { source: AuthorsError },
    #[snafu(display("Failed to render author data."))]
    Render { source: AsXmlError },
}
impl HttpStatus for AuthorsError {
    fn status_code(&self) -> StatusCode {
        match self {
            AuthorsError::Data { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            AuthorsError::Render { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(AuthorsError);

/// Render all authors as OPDS entries embedded in a feed.
#[utoipa::path(
    get,
    path = "/authors",
    tag = TAG,
    responses(
        (status = OK, content_type = "application/atom+xml"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler(State(state): State<Arc<AppState>>) -> Result<Response, AuthorsError> {
    let authors: Vec<DbAuthor> = state
        .calibre
        .authors(u32::MAX.into(), None, &SortOrder::ASC.into())
        .context(DataSnafu)?;

    let entries: Vec<Entry> = authors.into_iter().map(Entry::from).collect();
    let now = OffsetDateTime::now_utc();

    let self_link = Link {
        href: "/opds/authors".to_string(),
        media_type: MediaType::Navigation,
        rel: Relation::Myself,
        title: None,
        count: None,
    };
    let feed = Feed::create(
        now,
        &format!("{APP_NAME}:authors"),
        "All Authors",
        self_link,
        vec![],
        entries,
    );
    let xml = feed.as_xml().context(RenderSnafu)?;

    Ok(([(header::CONTENT_TYPE, "application/atom+xml")], xml).into_response())
}

#[derive(Debug, Snafu)]
pub enum SingleError {
    #[snafu(display("Failed to fetch author data."))]
    AuthorData { source: SingleAuthorError },
    #[snafu(display("Failed to render feed."))]
    FeedRender { source: AsXmlError },
}
impl HttpStatus for SingleError {
    fn status_code(&self) -> StatusCode {
        match self {
            SingleError::AuthorData { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            SingleError::FeedRender { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(SingleError);

/// Render a single series as an OPDS entry embedded in a feed.
#[utoipa::path(
    get,
    path = "/authors/{id}",
    tag = TAG,
    responses(
        (status = OK, content_type = "application/atom+xml"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn single(
    Path(id): Path<u64>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, SingleError> {
    let (author, books) = authors::single(id, &state.calibre, &state.config.library_path)
        .await
        .context(AuthorDataSnafu)?;

    let entries: Vec<Entry> = books.into_iter().map(Entry::from).collect();
    let now = OffsetDateTime::now_utc();

    let self_link = Link {
        href: format!("/opds/authors/{}", author.id),
        media_type: MediaType::Navigation,
        rel: Relation::Myself,
        title: None,
        count: None,
    };
    let feed = Feed::create(
        now,
        &format!("{APP_NAME}author:{}", author.id),
        &author.name,
        self_link,
        vec![],
        entries,
    );
    let xml = feed.as_xml().context(FeedRenderSnafu)?;

    Ok(([(header::CONTENT_TYPE, "application/atom+xml")], xml).into_response())
}
