use std::sync::Arc;

use crate::{
    APP_NAME,
    api::{
        SortOrder, TAG,
        error::{ErrorResponse, HttpStatus},
    },
    app_state::AppState,
    data::book::Book,
    http_error,
    opds::{
        entry::Entry, error::AsXmlError, feed::Feed, link::Link, media_type::MediaType,
        relation::Relation,
    },
};
use axum::{
    extract::State,
    http::{StatusCode, header},
    response::{IntoResponse, Response},
};
use calibre_db::calibre::BooksError;
use snafu::{ResultExt, Snafu};
use time::OffsetDateTime;

#[derive(Debug, Snafu)]
pub enum OdpsBooksError {
    #[snafu(display("Failed to fetch book data."))]
    Data { source: BooksError },
    #[snafu(display("Failed to render book data."))]
    Render { source: RenderError },
}
impl HttpStatus for OdpsBooksError {
    fn status_code(&self) -> StatusCode {
        match self {
            OdpsBooksError::Data { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            OdpsBooksError::Render { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(OdpsBooksError);

/// Render all books as OPDS entries embedded in a feed.
#[utoipa::path(
    get,
    path = "/books",
    tag = TAG,
    responses(
        (status = OK, content_type = "application/atom+xml"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler(State(state): State<Arc<AppState>>) -> Result<Response, OdpsBooksError> {
    let books: Vec<Book> = state
        .calibre
        .books(u32::MAX.into(), None, &SortOrder::ASC.into())
        .map(|x| {
            x.iter()
                .filter_map(|y| Book::full_book(y, &state.calibre, &state.config.library_path))
                .collect()
        })
        .context(DataSnafu)?;
    render_books(books).await.context(RenderSnafu)
}

#[derive(Debug, Snafu)]
pub enum RenderError {
    #[snafu(display("Failed to create opds feed."))]
    Feed { source: AsXmlError },
}

/// Render a list of books as OPDS entries in a feed.
pub(crate) async fn render_books(books: Vec<Book>) -> Result<Response, RenderError> {
    let entries: Vec<Entry> = books.into_iter().map(Entry::from).collect();
    let now = OffsetDateTime::now_utc();

    let self_link = Link {
        href: "/opds/books".to_string(),
        media_type: MediaType::Navigation,
        rel: Relation::Myself,
        title: None,
        count: None,
    };
    let feed = Feed::create(
        now,
        &format!("{APP_NAME}:books"),
        "All Books",
        self_link,
        vec![],
        entries,
    );
    let xml = feed.as_xml().context(FeedSnafu)?;

    Ok(([(header::CONTENT_TYPE, "application/atom+xml")], xml).into_response())
}
