use std::sync::Arc;

use crate::{
    APP_NAME,
    api::{
        TAG,
        books::{self, RecentBooksError},
        error::{ErrorResponse, HttpStatus},
    },
    app_state::AppState,
    http_error,
    opds::{
        entry::Entry, error::AsXmlError, feed::Feed, link::Link, media_type::MediaType,
        relation::Relation,
    },
};
use axum::{
    extract::State,
    http::{StatusCode, header},
    response::{IntoResponse, Response},
};
use snafu::{ResultExt, Snafu};
use time::OffsetDateTime;

#[derive(Debug, Snafu)]
pub enum RecentError {
    #[snafu(display("Failed to fetch recent books."))]
    Data { source: RecentBooksError },
    #[snafu(display("Failed to render feed."))]
    Render { source: AsXmlError },
}
impl HttpStatus for RecentError {
    fn status_code(&self) -> StatusCode {
        match self {
            RecentError::Data { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            RecentError::Render { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(RecentError);

#[utoipa::path(
    get,
    path = "/recent",
    tag = TAG,
    responses(
        (status = OK, content_type = "application/atom+xml"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler(State(state): State<Arc<AppState>>) -> Result<Response, RecentError> {
    let recent_books = books::recent(&state.calibre, &state.config.library_path)
        .await
        .context(DataSnafu)?;
    let entries: Vec<Entry> = recent_books.into_iter().map(Entry::from).collect();
    let now = OffsetDateTime::now_utc();

    let self_link = Link {
        href: "/opds/recent".to_string(),
        media_type: MediaType::Navigation,
        rel: Relation::Myself,
        title: None,
        count: None,
    };
    let feed = Feed::create(
        now,
        &format!("{APP_NAME}:recentbooks"),
        "Recent Books",
        self_link,
        vec![],
        entries,
    );
    let xml = feed.as_xml().context(RenderSnafu)?;

    Ok(([(header::CONTENT_TYPE, "application/atom+xml")], xml).into_response())
}
