use std::sync::Arc;

use crate::{
    APP_NAME,
    api::{
        TAG,
        error::{ErrorResponse, HttpStatus},
        search::{self, SearchQueryError},
    },
    app_state::AppState,
    http_error,
    opds::{
        error::AsXmlError,
        search::{OpenSearchDescription, Url},
    },
};
use axum::{
    extract::{Query, State},
    http::{StatusCode, header},
    response::{IntoResponse, Response},
};
use serde::Deserialize;
use snafu::{ResultExt, Snafu};

use super::books::{RenderError, render_books};

#[derive(Debug, Snafu)]
pub enum SearchError {
    #[snafu(display("Failed to query books."))]
    Query { source: SearchQueryError },
    #[snafu(display("Failed to render feed."))]
    Render { source: RenderError },
}
impl HttpStatus for SearchError {
    fn status_code(&self) -> StatusCode {
        match self {
            SearchError::Query { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            SearchError::Render { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(SearchError);

#[derive(Deserialize)]
pub struct Params {
    /// Query for a search request.
    query: String,
}
#[utoipa::path(
    get,
    path = "/search",
    tag = TAG,
    responses(
        (status = 200, content_type = "application/atom+xml"),
        (status = 500, description = "Error retrieving books from database", body = ErrorResponse)
    )
)]
pub async fn handler(
    Query(params): Query<Params>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, SearchError> {
    let books = search::query(&params.query, &state.calibre, &state.config.library_path)
        .await
        .context(QuerySnafu)?;

    render_books(books).await.context(RenderSnafu)
}

#[derive(Debug, Snafu)]
pub enum InfoError {
    #[snafu(display("Failed to render feed."))]
    FeedRender { source: AsXmlError },
}
impl HttpStatus for InfoError {
    fn status_code(&self) -> StatusCode {
        match self {
            InfoError::FeedRender { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}

http_error!(InfoError);

#[utoipa::path(
    get,
    path = "/search/info",
    tag = TAG,
    responses(
        (status = 200, content_type = "application/atom+xml"),
        (status = 500, description = "Internal error", body = ErrorResponse)
    )
)]
pub async fn info() -> Result<Response, InfoError> {
    let search = OpenSearchDescription {
        short_name: APP_NAME.to_string(),
        description: "Search for ebooks".to_string(),
        input_encoding: "UTF-8".to_string(),
        output_encoding: "UTF-8".to_string(),
        url: Url {
            type_name: "application/atom+xml".to_string(),
            template: "/opds/search?query={searchTerms}".to_string(),
        },
    };
    let xml = search.as_xml().context(FeedRenderSnafu)?;

    Ok(([(header::CONTENT_TYPE, "application/atom+xml")], xml).into_response())
}
