use std::sync::Arc;

use crate::{
    APP_NAME,
    api::{
        SortOrder, TAG,
        error::{ErrorResponse, HttpStatus},
        series::{self, SingleSeriesError},
    },
    app_state::AppState,
    http_error,
    opds::{
        entry::Entry, error::AsXmlError, feed::Feed, link::Link, media_type::MediaType,
        relation::Relation,
    },
};
use axum::{
    extract::{Path, State},
    http::{StatusCode, header},
    response::{IntoResponse, Response},
};
use calibre_db::calibre::MultipleSeriesError;
use snafu::{ResultExt, Snafu};
use time::OffsetDateTime;

#[derive(Debug, Snafu)]
pub enum SeriesError {
    #[snafu(display("Failed to fetch series data."))]
    Data { source: MultipleSeriesError },
    #[snafu(display("Failed to render series data."))]
    Render { source: AsXmlError },
}
impl HttpStatus for SeriesError {
    fn status_code(&self) -> StatusCode {
        match self {
            SeriesError::Data { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            SeriesError::Render { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(SeriesError);

/// Render all series as OPDS entries embedded in a feed.
#[utoipa::path(
    get,
    path = "/series",
    tag = TAG,
    responses(
        (status = OK, content_type = "application/atom+xml"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler(State(state): State<Arc<AppState>>) -> Result<Response, SeriesError> {
    let series = state
        .calibre
        .series(u32::MAX.into(), None, &SortOrder::ASC.into())
        .context(DataSnafu)?;

    let entries: Vec<Entry> = series.into_iter().map(Entry::from).collect();
    let now = OffsetDateTime::now_utc();

    let self_link = Link {
        href: "/opds/series".to_string(),
        media_type: MediaType::Navigation,
        rel: Relation::Myself,
        title: None,
        count: None,
    };
    let feed = Feed::create(
        now,
        &format!("{APP_NAME}:series"),
        "All Series",
        self_link,
        vec![],
        entries,
    );
    let xml = feed.as_xml().context(RenderSnafu)?;

    Ok(([(header::CONTENT_TYPE, "application/atom+xml")], xml).into_response())
}

#[derive(Debug, Snafu)]
pub enum SingleError {
    #[snafu(display("Failed to fetch series data."))]
    SeriesData { source: SingleSeriesError },
    #[snafu(display("Failed to render feed."))]
    FeedRender { source: AsXmlError },
}
impl HttpStatus for SingleError {
    fn status_code(&self) -> StatusCode {
        match self {
            SingleError::SeriesData { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            SingleError::FeedRender { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(SingleError);

/// Render a single series as an OPDS entry embedded in a feed.
#[utoipa::path(
    get,
    path = "/series/{id}",
    tag = TAG,
    responses(
        (status = OK, content_type = "application/atom+xml"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn single(
    Path(id): Path<u64>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, SingleError> {
    let (series, books) = series::single(id, &state.calibre, &state.config.library_path)
        .await
        .context(SeriesDataSnafu)?;

    let entries: Vec<Entry> = books.into_iter().map(Entry::from).collect();
    let now = OffsetDateTime::now_utc();

    let self_link = Link {
        href: format!("/opds/series/{}", series.id),
        media_type: MediaType::Navigation,
        rel: Relation::Myself,
        title: None,
        count: None,
    };
    let feed = Feed::create(
        now,
        &format!("{APP_NAME}:series:{}", series.id),
        &series.name,
        self_link,
        vec![],
        entries,
    );
    let xml = feed.as_xml().context(FeedRenderSnafu)?;

    Ok(([(header::CONTENT_TYPE, "application/atom+xml")], xml).into_response())
}
